@@ -3,6 +3,7 @@
 /* 
  * Copyright (C) 2001 Havoc Pennington
  * Copyright (C) 2002, 2003 Red Hat, Inc.
+ * Copyright (C) 2004 Rob Adams
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -31,7 +32,23 @@
 
 #include <X11/Xatom.h>
 
+#define WINDOW_HAS_TRANSIENT_TYPE(w)                    \
+          (w->type == META_WINDOW_DIALOG ||             \
+	   w->type == META_WINDOW_MODAL_DIALOG ||       \
+           w->type == META_WINDOW_TOOLBAR ||            \
+           w->type == META_WINDOW_MENU ||               \
+           w->type == META_WINDOW_UTILITY)
+
+#define WINDOW_TRANSIENT_FOR_WHOLE_GROUP(w)             \
+         ((w->xtransient_for == None ||                 \
+           w->transient_parent_is_root_window) &&       \
+          WINDOW_HAS_TRANSIENT_TYPE (w))
+
+#define WINDOW_IN_STACK(w) (w->stack_position >= 0)
+
 static void meta_stack_sync_to_server (MetaStack *stack);
+static void meta_window_set_stack_position_no_sync (MetaWindow *window,
+                                                    int         position);
 
 MetaStack*
 meta_stack_new (MetaScreen *screen)
@@ -107,8 +124,8 @@ meta_stack_remove (MetaStack  *stack,
   /* Set window to top position, so removing it will not leave gaps
    * in the set of positions
    */
-  meta_window_set_stack_position (window,
-                                  stack->n_positions - 1);
+  meta_window_set_stack_position_no_sync (window,
+                                          stack->n_positions - 1);
   window->stack_position = -1;
   stack->n_positions -= 1;  
 
@@ -147,8 +164,8 @@ void
 meta_stack_raise (MetaStack  *stack,
                   MetaWindow *window)
 {  
-  meta_window_set_stack_position (window,
-                                  stack->n_positions - 1);
+  meta_window_set_stack_position_no_sync (window,
+                                          stack->n_positions - 1);
   
   meta_stack_sync_to_server (stack);
 }
@@ -157,7 +174,7 @@ void
 meta_stack_lower (MetaStack  *stack,
                   MetaWindow *window)
 {
-  meta_window_set_stack_position (window, 0);
+  meta_window_set_stack_position_no_sync (window, 0);
   
   meta_stack_sync_to_server (stack);
 }
@@ -193,7 +210,9 @@ window_is_fullscreen_size (MetaWindow *window)
       
       meta_window_get_work_area_current_xinerama (window, &workarea);
       if (window->rect.x <= workarea.x &&
-          window->rect.y <= workarea.y)
+          window->rect.y <= workarea.y &&
+	  window->rect.x + window->rect.width >= workarea.x + workarea.width &&
+	  window->rect.y + window->rect.height >= workarea.y + workarea.height) 
         return TRUE;
     }
   
@@ -207,7 +226,9 @@ window_is_fullscreen_size (MetaWindow *window)
           
           meta_window_get_work_area_current_xinerama (window, &workarea);
           if (window->rect.x <= workarea.x &&
-              window->rect.y <= workarea.y) 
+              window->rect.y <= workarea.y &&
+	      window->rect.x + window->rect.width >= workarea.x + workarea.width &&
+	      window->rect.y + window->rect.height >= workarea.y + workarea.height) 
             return TRUE;
         }
       
@@ -217,11 +238,25 @@ window_is_fullscreen_size (MetaWindow *window)
   return FALSE;
 }
 
+static gboolean
+is_focused_foreach (MetaWindow *window,
+                    void       *data)
+{
+  if (window->has_focus ||
+      (window == window->display->expected_focus_window))
+    {
+      *((gboolean*) data) = TRUE;
+      return FALSE;
+    }
+  return TRUE;
+}
+
 /* Get layer ignoring any transient or group relationships */
 static MetaStackLayer
 get_standalone_layer (MetaWindow *window)
 {
   MetaStackLayer layer;
+  gboolean focused_transient = FALSE;
   
   switch (window->type)
     {
@@ -242,15 +277,15 @@ get_standalone_layer (MetaWindow *window)
       break;
       
     default:       
+      meta_window_foreach_transient (window,
+                                     is_focused_foreach,
+                                     &focused_transient);
 
-#if 0
-      if (window->has_focus &&
-          meta_prefs_get_focus_mode () == META_FOCUS_MODE_CLICK)
-        layer = META_LAYER_FOCUSED_WINDOW;
-#endif
-      
-      if (window->has_focus &&
-          (window->fullscreen || window_is_fullscreen_size (window)))
+      if (window->wm_state_below)
+        layer = META_LAYER_BOTTOM;
+      else if ((window->has_focus || focused_transient ||
+                (window == window->display->expected_focus_window)) &&
+               (window->fullscreen || window_is_fullscreen_size (window)))
         layer = META_LAYER_FULLSCREEN;
       else if (window->wm_state_above)
         layer = META_LAYER_DOCK;
@@ -262,36 +297,6 @@ get_standalone_layer (MetaWindow *window)
   return layer;
 }
 
-typedef struct
-{
-  MetaStackLayer max;
-} MaxLayerData;
-
-static gboolean
-max_layer_func (MetaWindow *window,
-                void       *data)
-{
-  MaxLayerData *d = data;
-  MetaStackLayer layer;
-  
-  layer = get_standalone_layer (window);
-  if (layer > d->max)
-    d->max = layer;
-
-  return TRUE;
-}
-
-static MetaStackLayer
-get_maximum_layer_of_ancestor (MetaWindow *window)
-{
-  MaxLayerData d;
-
-  d.max = get_standalone_layer (window);
-  meta_window_foreach_ancestor (window, max_layer_func, &d);
-  
-  return d.max;
-}
-
 /* Note that this function can never use window->layer only
  * get_standalone_layer, or we'd have issues.
  */
@@ -341,58 +346,29 @@ compute_layer (MetaWindow *window)
    * windows getting in fullscreen layer if any terminal is
    * fullscreen.
    */
-  if (window->type == META_WINDOW_DIALOG ||
-      window->type == META_WINDOW_MODAL_DIALOG ||
-      window->type == META_WINDOW_UTILITY ||
-      window->type == META_WINDOW_MENU ||
-      window->type == META_WINDOW_TOOLBAR)
+  if (WINDOW_HAS_TRANSIENT_TYPE(window) &&
+      (window->xtransient_for == None ||
+       window->transient_parent_is_root_window))
     {
-      if (window->xtransient_for != None &&
-          !window->transient_parent_is_root_window)
+      /* We only do the group thing if the dialog is NOT transient for
+       * a particular window. Imagine a group with a normal window, a dock,
+       * and a dialog transient for the normal window; you don't want the dialog
+       * above the dock if it wouldn't normally be.
+       */
+      
+      MetaStackLayer group_max;
+      
+      group_max = get_maximum_layer_in_group (window);
+      
+      if (group_max > window->layer)
         {
-          MetaStackLayer ancestor_max;
-
-          ancestor_max = get_maximum_layer_of_ancestor (window);
-          
-          if (ancestor_max > window->layer)
-            {
-              meta_topic (META_DEBUG_STACK,
-                          "Promoting window %s from layer %d to %d due to transiency\n",
-                          window->desc, window->layer, ancestor_max);
-              window->layer = ancestor_max;
-            }
-        }
-      else
-        {
-          /* We only do the group thing if the dialog is NOT transient for
-           * a particular window. Imagine a group with a normal window, a dock,
-           * and a dialog transient for the normal window; you don't want the dialog
-           * above the dock if it wouldn't normally be.
-           */
-
-          /* FIXME when promoting a window here,
-           * it's necessary to promote its transient children
-           * (or other windows constrained to be above it)
-           * as well, but we aren't handling that, and it's
-           * somewhat hard to fix.
-           *
-           * http://bugzilla.gnome.org/show_bug.cgi?id=96140
-           */
-          
-          MetaStackLayer group_max;
-          
-          group_max = get_maximum_layer_in_group (window);
-          
-          if (group_max > window->layer)
-            {
-              meta_topic (META_DEBUG_STACK,
-                          "Promoting window %s from layer %d to %d due to group membership\n",
-                          window->desc, window->layer, group_max);
-              window->layer = group_max;
-            }
+          meta_topic (META_DEBUG_STACK,
+                      "Promoting window %s from layer %d to %d due to group membership\n",
+                      window->desc, window->layer, group_max);
+          window->layer = group_max;
         }
     }
-  
+
   meta_topic (META_DEBUG_STACK, "Window %s on layer %d type = %d has_focus = %d\n",
               window->desc, window->layer,
               window->type, window->has_focus);
@@ -522,20 +498,6 @@ add_constraint (Constraint **constraints,
   constraints[below->stack_position] = c;
 }
 
-#define WINDOW_HAS_TRANSIENT_TYPE(w)                    \
-          (w->type == META_WINDOW_DIALOG ||             \
-	   w->type == META_WINDOW_MODAL_DIALOG ||       \
-           w->type == META_WINDOW_TOOLBAR ||            \
-           w->type == META_WINDOW_MENU ||               \
-           w->type == META_WINDOW_UTILITY)
-
-#define WINDOW_TRANSIENT_FOR_WHOLE_GROUP(w)             \
-         ((w->xtransient_for == None ||                 \
-           w->transient_parent_is_root_window) &&       \
-          WINDOW_HAS_TRANSIENT_TYPE (w))
-
-#define WINDOW_IN_STACK(w) (w->stack_position >= 0)
-
 static void
 create_constraints (Constraint **constraints,
                     GList       *windows)
@@ -697,13 +659,21 @@ static void
 ensure_above (MetaWindow *above,
               MetaWindow *below)
 {  
+  if (WINDOW_HAS_TRANSIENT_TYPE(above) &&
+      above->layer < below->layer)
+    {
+      meta_topic (META_DEBUG_STACK,
+		  "Promoting window %s from layer %d to %d due to contraint\n",
+		  above->desc, above->layer, below->layer);
+      above->layer = below->layer;
+    }
+
   if (above->stack_position < below->stack_position)
     {
       /* move above to below->stack_position bumping below down the stack */
-      meta_window_set_stack_position (above, below->stack_position);
+      meta_window_set_stack_position_no_sync (above, below->stack_position);
       g_assert (below->stack_position + 1 == above->stack_position);
     }
-
   meta_topic (META_DEBUG_STACK, "%s above at %d > %s below at %d\n",
               above->desc, above->stack_position,
               below->desc, below->stack_position);
@@ -917,6 +887,7 @@ meta_stack_ensure_sorted (MetaStack *stack)
                           w->desc, old_layer, w->layer);
               
               stack->need_resort = TRUE;
+              stack->need_constrain = TRUE;
               /* don't need to constrain as constraining
                * purely operates in terms of stack_position
                * not layer
@@ -929,7 +900,8 @@ meta_stack_ensure_sorted (MetaStack *stack)
       stack->need_relayer = FALSE;
     }
 
-  /* Update stack_position to reflect transiency constraints */
+  /* Update stack_position and layer to reflect transiency
+     constraints */
   constrain_stacking (stack);
   
   /* Sort stack->sorted with layers having priority over stack_position
@@ -1372,7 +1344,7 @@ get_default_focus_window (MetaStack     *stack,
           !window->minimized &&
           (window->input || window->take_focus) &&
           (workspace == NULL ||
-           meta_window_visible_on_workspace (window, workspace)))
+           meta_window_located_on_workspace (window, workspace)))
         {
           if (topmost_dock == NULL &&
               window->type == META_WINDOW_DOCK)
@@ -1459,7 +1431,7 @@ meta_stack_list_windows (MetaStack     *stack,
       MetaWindow *window = link->data;
       
       if (window &&
-          (workspace == NULL || meta_window_visible_on_workspace (window, workspace)))
+          (workspace == NULL || meta_window_located_on_workspace (window, workspace)))
         {
           workspace_windows = g_list_prepend (workspace_windows,
                                               window);
@@ -1495,8 +1467,8 @@ meta_stack_windows_cmp  (MetaStack  *stack,
 }
 
 void
-meta_window_set_stack_position (MetaWindow *window,
-                                int         position)
+meta_window_set_stack_position_no_sync (MetaWindow *window,
+                                        int         position)
 {
   int low, high, delta;
   GList *tmp;
@@ -1547,3 +1519,11 @@ meta_window_set_stack_position (MetaWindow *window,
               "Window %s had stack_position set to %d\n",
               window->desc, window->stack_position);
 }
+
+void
+meta_window_set_stack_position (MetaWindow *window,
+                                int         position)
+{
+  meta_window_set_stack_position_no_sync (window, position);
+  meta_stack_sync_to_server (window->screen->stack);
+}
